from inewave.nwlistop.modelos.blocos.ree import REE
from inewave.nwlistop.modelos.arquivos.arquivoree import ArquivoREE
from inewave.nwlistop.modelos.evert import EvertAnos


class Evert(ArquivoREE):
    """
    Armazena os dados das saídas referentes ao vertimento de reservatórios
    , por REE.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `evert00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        REE,
        EvertAnos,
    ]



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
)
from inewave.nwlistop.modelos.eafbm import EafsAnos


class Eafbm(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes às energias
    afluentes brutas, por submercado em valores absolutos.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `eafbm00x.out`, onde x varia conforme o
    submercado em questão.

    """

    BLOCKS = [
        Submercado,
        EafsAnos,
    ]



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.ghiduh import GhidAnos


class Ghiduh(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes à geração hidráulica por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `ghiduh00x.out`, onde x varia conforme a
    usina em questão.

    """

    BLOCKS = [
        Usina,
        GhidAnos,
    ]



# Inclui os membros

from .nwlistopdat import Nwlistopdat  # noqa
from .cdef import Cdef  # noqa
from .cdefsin import Cdefsin  # noqa
from .cmarg import Cmarg  # noqa
from .cmargmed import Cmargmed  # noqa
from .coper import Coper  # noqa
from .corteolm import Corteolm  # noqa
from .cterm import Cterm  # noqa
from .ctermsin import Ctermsin  # noqa
from .deficit import Def  # noqa
from .defsin import Defsin  # noqa
from .depminuh import Depminuh  # noqa
from .dfphauh import Dfphauh  # noqa
from .dlppdfmax import Dlppdfmax  # noqa
from .dlppdfmaxs import Dlppdfmaxs  # noqa
from .dlpptbmax import Dlpptbmax  # noqa
from .dlpptbmaxm import Dlpptbmaxm  # noqa
from .dlpptbmaxs import Dlpptbmaxs  # noqa
from .dtbmax import Dtbmax  # noqa
from .dtbmin import Dtbmin  # noqa
from .dvazmax import Dvazmax  # noqa
from .evert import Evert  # noqa
from .evertm import Evertm  # noqa
from .evertsin import Evertsin  # noqa
from .exces import Exces  # noqa
from .excessin import Excessin  # noqa
from .fteolm import Fteolm  # noqa
from .fteolsin import Fteolsin  # noqa
from .eaf import Eaf  # noqa
from .eafb import Eafb  # noqa
from .eafbm import Eafbm  # noqa
from .eafbsin import Eafbsin  # noqa
from .eafm import Eafm  # noqa
from .earmf import Earmf  # noqa
from .earmfm import Earmfm  # noqa
from .earmfp import Earmfp  # noqa
from .earmfpm import Earmfpm  # noqa
from .earmfsin import Earmfsin  # noqa
from .earmfpsin import Earmfpsin  # noqa
from .geol import Geol  # noqa
from .geolm import Geolm  # noqa
from .geolsin import Geolsin  # noqa
from .ghiduh import Ghiduh  # noqa
from .ghmax import Ghmax  # noqa
from .ghmaxm import Ghmaxm  # noqa
from .ghmaxmr import Ghmaxmr  # noqa
from .ghmaxr import Ghmaxr  # noqa
from .ghmaxrsin import Ghmaxrsin  # noqa
from .ghmaxsin import Ghmaxsin  # noqa
from .ghtot import Ghtot  # noqa
from .ghtotm import Ghtotm  # noqa
from .ghtotsin import Ghtotsin  # noqa
from .gtert import Gtert  # noqa
from .gttot import Gttot  # noqa
from .gttotsin import Gttotsin  # noqa
from .intercambio import Intercambio  # noqa
from .invade import Invade  # noqa
from .invadem import Invadem  # noqa
from .mediasmerc import Mediasmerc  # noqa
from .mediassin import Mediassin  # noqa
from .mercl import Mercl  # noqa
from .merclsin import Merclsin  # noqa
from .perdf import Perdf  # noqa
from .perdfm import Perdfm  # noqa
from .perdfsin import Perdfsin  # noqa
from .qafluh import Qafluh  # noqa
from .qincruh import Qincruh  # noqa
from .rhslppdf import Rhslppdf  # noqa
from .rhslpptb import Rhslpptb  # noqa
from .vagua import Vagua  # noqa
from .varmpuh import Varmpuh  # noqa
from .varmuh import Varmuh  # noqa
from .vento import Vento  # noqa
from .vertuh import Vertuh  # noqa
from .verturb import Verturb  # noqa
from .verturbm import Verturbm  # noqa
from .verturbsin import Verturbsin  # noqa
from .vevmin import Vevmin  # noqa
from .vevminm import Vevminm  # noqa
from .vevminsin import Vevminsin  # noqa
from .vghmin import Vghmin  # noqa
from .vghminm import Vghminm  # noqa
from .vghminsin import Vghminsin  # noqa
from .vghminuh import Vghminuh  # noqa
from .vturuh import Vturuh  # noqa



from cfinterface.files.sectionfile import SectionFile
from inewave.newave.modelos.vazaof import SecaoDadosVazaof
import pandas as pd  # type: ignore

from typing import TypeVar, Optional


class Vazaof(SectionFile):
    """
    Armazena os dados de saída do NEWAVE referentes às séries sintéticas
    de vazão para a etapa forward geradas pelo modelo.
    """

    T = TypeVar("T")

    SECTIONS = [SecaoDadosVazaof]
    STORAGE = "BINARY"

    @property
    def series(self) -> Optional[pd.DataFrame]:
        """
        Obtém a tabela com os dados das séries de vazão
        afluente por UHE e por estágio.

        - estagio (`int`): estágio do cenário gerado
        - uhe (`int`): UHE para a qual foi gerado
        - serie (`int`): índice da série forward
        - valor (`float`): vazão em hm3

        :return: A tabela com os dados das séries
        :rtype: pd.DataFrame | None
        """
        sections = [r for r in self.data.of_type(SecaoDadosVazaof)]
        if len(sections) > 0:
            return sections[0].data
        else:
            return None

    @series.setter
    def series(self, df: pd.DataFrame):
        sections = [r for r in self.data.of_type(SecaoDadosVazaof)]
        if len(sections) > 0:
            sections[0].data = df



import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from datetime import datetime
from inewave.config import MESES_DF
from typing import List

__COLS_IDENTIFICACAO = ["data", "ano", "serie", "patamar", "classe"]


def __formata_df_meses_para_datas_serie(df: pd.DataFrame) -> pd.DataFrame:
    anos = np.array(df["ano"].unique().tolist())
    datas = pd.date_range(
        datetime(year=anos[0], month=1, day=1),
        datetime(year=anos[-1], month=12, day=1),
        freq="MS",
    )
    df_series = pd.DataFrame()
    for a in anos:
        df_ano = df.loc[df["ano"] == a, MESES_DF].T
        df_ano.columns = [str(s) for s in list(range(1, df_ano.shape[1] + 1))]
        df_series = pd.concat([df_series, df_ano], ignore_index=True)
    df_series["data"] = datas
    cols_not_scenarios = [
        c for c in df_series.columns if c in __COLS_IDENTIFICACAO
    ]
    cols_scenarios = [
        c for c in df_series.columns if c not in __COLS_IDENTIFICACAO
    ]
    df_formatado = pd.melt(
        df_series,
        id_vars=cols_not_scenarios,
        value_vars=cols_scenarios,
        var_name="serie",
        value_name="valor",
    )
    return df_formatado


def __formata_df_meses_para_datas_serie_patamar(
    df: pd.DataFrame,
) -> pd.DataFrame:
    anos = np.array(df["ano"].unique().tolist())
    patamares = np.array(df["patamar"].unique().tolist())
    datas = []
    for a in anos:
        for _ in patamares:
            datas += pd.date_range(
                datetime(year=a, month=1, day=1),
                datetime(year=a, month=12, day=1),
                freq="MS",
            ).tolist()
    df_series = pd.DataFrame()
    for a in anos:
        for p in patamares:
            df_ano_patamar = df.loc[
                (df["ano"] == a) & (df["patamar"] == p),
                MESES_DF,
            ].T
            cols = [
                str(s) for s in list(range(1, df_ano_patamar.shape[1] + 1))
            ]
            df_ano_patamar.columns = cols
            df_ano_patamar["patamar"] = str(p)
            df_ano_patamar = df_ano_patamar[["patamar"] + cols]
            df_series = pd.concat(
                [df_series, df_ano_patamar], ignore_index=True
            )
    df_series["data"] = datas
    cols_not_scenarios = [
        c for c in df_series.columns if c in __COLS_IDENTIFICACAO
    ]
    cols_scenarios = [
        c for c in df_series.columns if c not in __COLS_IDENTIFICACAO
    ]
    df_formatado = pd.melt(
        df_series,
        id_vars=cols_not_scenarios,
        value_vars=cols_scenarios,
        var_name="serie",
        value_name="valor",
    )
    return df_formatado[["data", "patamar", "serie", "valor"]]


def __formata_df_meses_para_datas_classetermica_serie_patamar(
    df: pd.DataFrame,
) -> pd.DataFrame:
    anos = np.array(df["ano"].unique().tolist())
    patamares = np.array(df["patamar"].unique().tolist())
    classes = np.array(df["classe"].unique().tolist())
    datas = []
    for a in anos:
        for _ in patamares:
            for _ in classes:
                datas += pd.date_range(
                    datetime(year=a, month=1, day=1),
                    datetime(year=a, month=12, day=1),
                    freq="MS",
                ).tolist()
    df_series = pd.DataFrame()
    for a in anos:
        for c in classes:
            for p in patamares:
                df_ano_classe_patamar = df.loc[
                    (df["ano"] == a)
                    & (df["classe"] == c)
                    & (df["patamar"] == p),
                    MESES_DF,
                ].T
                cols = [
                    str(s)
                    for s in list(range(1, df_ano_classe_patamar.shape[1] + 1))
                ]
                df_ano_classe_patamar.columns = cols
                df_ano_classe_patamar["classe"] = str(c)
                df_ano_classe_patamar["patamar"] = str(p)
                df_ano_classe_patamar = df_ano_classe_patamar[
                    ["classe", "patamar"] + cols
                ]
                df_series = pd.concat(
                    [df_series, df_ano_classe_patamar], ignore_index=True
                )
    df_series["data"] = datas
    cols_not_scenarios = [
        c for c in df_series.columns if c in __COLS_IDENTIFICACAO
    ]
    cols_scenarios = [
        c for c in df_series.columns if c not in __COLS_IDENTIFICACAO
    ]
    df_formatado = pd.melt(
        df_series,
        id_vars=cols_not_scenarios,
        value_vars=cols_scenarios,
        var_name="serie",
        value_name="valor",
    )
    return df_formatado[["classe", "data", "patamar", "serie", "valor"]]


def formata_df_meses_para_datas_nwlistop(df: pd.DataFrame) -> pd.DataFrame:
    colunas_df = df.columns.tolist()
    colunas_identificacao = tuple(
        [c for c in colunas_df if c in __COLS_IDENTIFICACAO]
    )
    mapa_formatacao = {
        ("ano", "serie"): __formata_df_meses_para_datas_serie,
        (
            "ano",
            "serie",
            "patamar",
        ): __formata_df_meses_para_datas_serie_patamar,
        (
            "ano",
            "classe",
            "serie",
            "patamar",
        ): __formata_df_meses_para_datas_classetermica_serie_patamar,
    }
    return mapa_formatacao[colunas_identificacao](df)


def prepara_vetor_anos_tabela(anos: List[str]) -> List[datetime]:
    # Se tem pré, substitui por 0001
    # Se tem pós, substitui por 9999
    # Repete os valores existentes 12 vezes
    anos_convertidos: List[int] = []
    for a in anos:
        if a == "PRE":
            a_convertido = 1
        elif a == "POS":
            a_convertido = 9999
        else:
            a_convertido = int(a)
        anos_convertidos.append(a_convertido)

    anos_array = np.array(anos_convertidos).repeat(len(MESES_DF))
    meses = np.tile(np.arange(1, 13), len(anos))
    return [
        datetime(year=a, month=m, day=1) for a, m in zip(anos_array, meses)
    ]


def repete_vetor(
    valores: list, n_repeticoes: int = len(MESES_DF)
) -> np.ndarray:
    return np.array(valores).repeat(n_repeticoes)


def prepara_valor_ano(ano: int) -> str:
    if ano == 1:
        return "PRE"
    elif ano == 9999:
        return "POS"
    else:
        return str(ano)
